import os
import copy
import logging
import datetime
from itertools import product
from collections import defaultdict
from typing import Callable, Dict, List, Tuple, Union

import csv
import h5py
import numpy as np
import pandas as pd

from ml4h.defines import CARDIAC_SURGERY_DATE_FORMAT
from ml4h.normalizer import Standardize, ZeroMeanStd1
from ml4h.defines import ECG_REST_AMP_LEADS, ECG_REST_UKB_LEADS
from ml4h.defines import PARTNERS_DATE_FORMAT, PARTNERS_DATETIME_FORMAT
from ml4h.TensorMap import TensorMap, str2date, Interpretation, decompress_data
from ml4h.tensormap.mgb.ecg import _get_ecg_dates, _is_dynamic_shape, _make_hd5_path, make_voltage
from ml4h.tensormap.mgb.ecg import validator_not_all_zero, _hd5_filename_to_mrn_int, _resample_voltage

YEAR_DAYS = 365.26
INCIDENCE_CSV = '/media/erisone_snf13/lc_outcomes.csv'
CARDIAC_SURGERY_OUTCOMES_CSV = '/data/sts-data/mgh-preop-ecg-outcome-labels.csv'
PARTNERS_PREFIX = 'partners_ecg_rest'
WIDE_FILE = '/home/sam/ml/hf-wide-2020-09-15-with-lvh-and-lbbb.tsv'
#WIDE_FILE = '/home/sam/ml/mgh-wide-2020-06-25-with-mrn.tsv'


def make_mgb_dynamic_tensor_maps(desired_map_name: str) -> TensorMap:
    tensor_map_maker_fxns = [
        make_waveform_maps, make_partners_diagnosis_maps, make_waveform_maps_for_ukb, make_waveform_maps_lead_I,
    ]
    for map_maker_function in tensor_map_maker_fxns:
        desired_map = map_maker_function(desired_map_name)
        if desired_map is not None:
            return desired_map


def make_waveform_maps(desired_map_name: str) -> TensorMap:
    """Creates 12 possible Tensor Maps and returns the desired one or None:

        mgb_ecg_2500_std    mgb_ecg_2500_std_exact   mgb_ecg_5000_std    mgb_ecg_5000_std_exact
        mgb_ecg_2500_mv     mgb_ecg_2500_mv_exact    mgb_ecg_5000_mv     mgb_ecg_5000_mv_exact
        mgb_ecg_2500_raw    mgb_ecg_2500_raw_exact   mgb_ecg_5000_raw    mgb_ecg_5000_raw_exact

        _std normalizes with ZeroMeanStd1 and resamples
        _mv normalizes with Standardize mean = 0, std = 1000
        _raw does not normalize
        _exact does not resample
    :param desired_map_name: The name of the TensorMap and
    :return: The desired TensorMap
    """
    length_options = [2500, 5000]
    exact_options = [True, False]
    normalize_options = [ZeroMeanStd1(), Standardize(mean=0, std=1000), None]
    for length, exact_length, normalization in product(length_options, exact_options, normalize_options):
        norm = '_std' if isinstance(normalization, ZeroMeanStd1) else '_mv' if isinstance(normalization, Standardize) else '_raw'
        exact = '_exact' if exact_length else ''
        name = f'ecg_{length}{norm}{exact}'
        #name = f'ecg_{length}_std'
        if name == desired_map_name:
            return TensorMap(
                name,
                shape=(length, 12),
                path_prefix=PARTNERS_PREFIX,
                tensor_from_file=make_voltage(exact_length),
                normalization=normalization,
                channel_map=ECG_REST_AMP_LEADS,
            )


def make_waveform_maps_lead_I(desired_map_name: str) -> TensorMap:
    """Creates 12 possible Tensor Maps and returns the desired one or None:
    :param desired_map_name: The name of the TensorMap and
    :return: The desired TensorMap
    """
    length_options = [2500, 5000]
    exact_options = [True, False]
    normalize_options = [ZeroMeanStd1(), Standardize(mean=0, std=1000), None]
    for length, exact_length, normalization in product(length_options, exact_options, normalize_options):
        norm = '_std' if isinstance(normalization, ZeroMeanStd1) else '_mv' if isinstance(normalization, Standardize) else '_raw'
        exact = '_exact' if exact_length else ''
        name = f'ecg_lead_I_{length}{norm}{exact}'
        if name == desired_map_name:
            return TensorMap(
                name,
                shape=(length, 1),
                path_prefix=PARTNERS_PREFIX,
                tensor_from_file=make_voltage(exact_length),
                normalization=normalization,
                channel_map={'I': 0},
            )


def _dummy_tensor_from_file(tm, hd5, dependents={}):
    return np.zeros(tm.shape, dtype=np.float32)

def make_waveform_maps_for_ukb(desired_map_name: str) -> TensorMap:
    """Creates Tensor Maps and returns the desired one or None
    :param desired_map_name: The name of the TensorMap and
    :return: The desired TensorMap
    """
    if 'strip' == desired_map_name:
        return TensorMap(
            desired_map_name,
            shape=(5000, 12),
            path_prefix=PARTNERS_PREFIX,
            tensor_from_file=make_voltage(False),
            normalization=ZeroMeanStd1(), #normalization,
            channel_map=ECG_REST_AMP_LEADS,
        )
    elif 'ecg_rest_median_raw_10' == desired_map_name:
        return TensorMap(
            desired_map_name,
            shape=(600, 12),
            tensor_from_file=_dummy_tensor_from_file,
            channel_map=ECG_REST_AMP_LEADS,
        )

def make_lead_maps(desired_map_name: str) -> TensorMap:
    for lead in ECG_REST_AMP_LEADS:
        tensormap_name = f'lead_{lead}_len'
        if desired_map_name == tensormap_name:
            return TensorMap(
                tensormap_name, interpretation=Interpretation.CATEGORICAL, path_prefix=PARTNERS_PREFIX,
                channel_map={'_2500': 0, '_5000': 1, 'other': 2}, time_series_limit=0, validator=validator_not_all_zero,
                tensor_from_file=make_voltage_len_categorical_tmap(lead=lead),
            )


def make_voltage_len_categorical_tmap(lead, channel_prefix='_', channel_unknown='other'):
    def _tensor_from_file(tm, hd5, dependents={}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        tensor = np.zeros(shape, dtype=float)
        for i, ecg_date in enumerate(ecg_dates):
            path = _make_hd5_path(tm, ecg_date, lead)
            try:
                lead_len = hd5[path].attrs['len']
                lead_len = f'{channel_prefix}{lead_len}'
                matched = False
                for cm in tm.channel_map:
                    if lead_len.lower() == cm.lower():
                        slices = (i, tm.channel_map[cm]) if dynamic else (tm.channel_map[cm],)
                        tensor[slices] = 1.0
                        matched = True
                        break
                if not matched:
                    slices = (i, tm.channel_map[channel_unknown]) if dynamic else (tm.channel_map[channel_unknown],)
                    tensor[slices] = 1.0
            except KeyError:
                logging.debug(f'Could not get voltage length for lead {lead} from ECG on {ecg_date} in {hd5.filename}')
        return tensor
    return _tensor_from_file


# Date formatting
def _partners_str2date(d) -> datetime.date:
    return datetime.datetime.strptime(d, PARTNERS_DATE_FORMAT).date()


def _loyalty_str2date(date_string: str) -> datetime.date:
    return str2date(date_string.split(' ')[0])


def _cardiac_surgery_str2date(input_date: str, date_format: str = CARDIAC_SURGERY_DATE_FORMAT) -> datetime.datetime:
    return datetime.datetime.strptime(input_date, date_format)


def build_incidence_tensor_from_file(
    file_name: str, patient_column: str = 'Mrn', birth_column: str = 'birth_date',
    diagnosis_column: str = 'first_stroke', start_column: str = 'start_fu',
    delimiter: str = ',', incidence_only: bool = False, check_birthday: bool = True,
) -> Callable:
    """Build a tensor_from_file function for future (and prior) diagnoses given a TSV of patients and diagnosis dates.

    The tensor_from_file function returned here should be used
    with CATEGORICAL TensorMaps to classify patients by disease state.

    :param file_name: CSV or TSV file with header of patient IDs (MRNs) dates of enrollment and dates of diagnosis
    :param patient_column: The header name of the column of patient ids
    :param diagnosis_column: The header name of the column of disease diagnosis dates
    :param start_column: The header name of the column of enrollment dates
    :param delimiter: The delimiter separating columns of the TSV or CSV
    :param incidence_only: Flag to skip patients whose diagnosis date is prior to acquisition date of input data
    :return: The tensor_from_file function to provide to TensorMap constructors
    """
    error = None
    try:
        with open(file_name, 'r', encoding='utf-8') as f:
            reader = csv.reader(f, delimiter=delimiter)
            header = next(reader)
            patient_index = header.index(patient_column)
            birth_index = header.index(birth_column)
            start_index = header.index(start_column)
            date_index = header.index(diagnosis_column)
            date_table = {}
            birth_table = {}
            patient_table = {}
            for row in reader:
                try:
                    patient_key = int(row[patient_index])
                    patient_table[patient_key] = _loyalty_str2date(row[start_index])
                    birth_table[patient_key] = _loyalty_str2date(row[birth_index])
                    if row[date_index] == '' or row[date_index] == 'NULL':
                        continue
                    date_table[patient_key] = _loyalty_str2date(row[date_index])
                    if len(patient_table) % 2000 == 0:
                        logging.debug(f'Processed: {len(patient_table)} patient rows.')
                except ValueError as e:
                    logging.warning(f'val err {e}')
            logging.info(f'Done processing {diagnosis_column} Got {len(patient_table)} patient rows and {len(date_table)} events.')
    except FileNotFoundError as e:
        error = e

    def tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        if error:
            raise error

        ecg_dates = _get_ecg_dates(tm, hd5)
        if len(ecg_dates) > 1:
            raise NotImplementedError('Diagnosis models for multiple ECGs are not implemented.')
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        categorical_data = np.zeros(shape, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            path = lambda key: _make_hd5_path(tm, ecg_date, key)
            mrn = _hd5_filename_to_mrn_int(hd5.filename)
            mrn_int = int(mrn)

            if mrn_int not in patient_table:
                raise KeyError(f'{tm.name} mrn not in incidence csv')

            if check_birthday:
                birth_date = _partners_str2date(decompress_data(data_compressed=hd5[path('dateofbirth')][()], dtype=hd5[path('dateofbirth')].attrs['dtype']))
                if birth_date != birth_table[mrn_int]:
                    raise ValueError(f'Birth dates do not match! CSV had {birth_table[patient_key]} but HD5 has {birth_date}')

            assess_date = _partners_str2date(decompress_data(data_compressed=hd5[path('acquisitiondate')][()], dtype=hd5[path('acquisitiondate')].attrs['dtype']))
            if assess_date < patient_table[mrn_int]:
                raise ValueError(f'{tm.name} Assessed earlier than enrollment')
            if mrn_int not in date_table:
                index = 0
            else:
                disease_date = date_table[mrn_int]

                if incidence_only and disease_date < assess_date:
                    raise ValueError(f'{tm.name} is skipping prevalent cases.')
                elif incidence_only and disease_date >= assess_date:
                    index = 1
                else:
                    index = 1 if disease_date < assess_date else 2
                logging.debug(f'mrn: {mrn_int}  Got disease_date: {disease_date} assess  {assess_date} index  {index}.')
            slices = (i, index) if dynamic else (index,)
            categorical_data[slices] = 1.0
        return categorical_data
    return tensor_from_file


def _diagnosis_channels(disease: str, incidence_only: bool = False):
    if incidence_only:
        return {f'no_{disease}': 0,  f'future_{disease}': 1}
    return {f'no_{disease}': 0, f'prior_{disease}': 1, f'future_{disease}': 2}


def _outcome_channels(outcome: str):
    return {f'no_{outcome}': 0,  f'{outcome}': 1}


def loyalty_time_to_event(
    file_name: str, incidence_only: bool = False, patient_column: str = 'Mrn',
    follow_up_start_column: str = 'start_fu', follow_up_total_column: str = 'total_fu',
    diagnosis_column: str = 'first_stroke', delimiter: str = ',',
):
    """Build a tensor_from_file function for modeling relative time to event of diagnoses given a TSV of patients and dates.

    The tensor_from_file function returned here should be used
    with TIME_TO_EVENT TensorMaps to model relative time free from a diagnosis for a given disease.

    :param file_name: CSV or TSV file with header of patient IDs (MRNs) dates of enrollment and dates of diagnosis
    :param incidence_only: Flag to skip patients whose diagnosis date is prior to acquisition date of input data
    :param patient_column: The header name of the column of patient ids
    :param follow_up_start_column: The header name of the column of enrollment dates
    :param follow_up_total_column: The header name of the column with total enrollment time (in years)
    :param diagnosis_column: The header name of the column of disease diagnosis dates
    :param delimiter: The delimiter separating columns of the TSV or CSV
    :return: The tensor_from_file function to provide to TensorMap constructors
    """
    error = None
    disease_dicts = defaultdict(dict)
    try:
        with open(file_name, 'r', encoding='utf-8') as f:
            reader = csv.reader(f, delimiter=delimiter)
            header = next(reader)
            follow_up_start_index = header.index(follow_up_start_column)
            follow_up_total_index = header.index(follow_up_total_column)
            patient_index = header.index(patient_column)
            date_index = header.index(diagnosis_column)
            for row in reader:
                try:
                    patient_key = int(row[patient_index])
                    disease_dicts['follow_up_start'][patient_key] = _loyalty_str2date(row[follow_up_start_index])
                    disease_dicts['follow_up_total'][patient_key] = float(row[follow_up_total_index])
                    if row[date_index] == '' or row[date_index] == 'NULL':
                        continue
                    disease_dicts['diagnosis_dates'][patient_key] = _loyalty_str2date(row[date_index])
                    if len(disease_dicts['follow_up_start']) % 2000 == 0:
                        logging.debug(f"Processed: {len(disease_dicts['follow_up_start'])} patient rows.")
                except ValueError as e:
                    logging.warning(f'val err {e}')
            logging.info(f"Done processing {diagnosis_column} Got {len(disease_dicts['follow_up_start'])} patient rows and {len(disease_dicts['diagnosis_dates'])} events.")
    except FileNotFoundError as e:
        error = e

    def _cox_tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        if error:
            raise error

        ecg_dates = _get_ecg_dates(tm, hd5)
        if len(ecg_dates) > 1:
            raise NotImplementedError('Cox hazard models for multiple ECGs are not implemented.')
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        tensor = np.zeros(tm.shape, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            patient_key_from_ecg = _hd5_filename_to_mrn_int(hd5.filename)
            if patient_key_from_ecg not in disease_dicts['follow_up_start']:
                raise KeyError(f'{tm.name} mrn not in incidence csv')

            path = _make_hd5_path(tm, ecg_date, 'acquisitiondate')
            assess_date = _partners_str2date(decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype']))
            if assess_date < disease_dicts['follow_up_start'][patient_key_from_ecg]:
                raise ValueError(f'Assessed earlier than enrollment.')

            if patient_key_from_ecg not in disease_dicts['diagnosis_dates']:
                has_disease = 0
                censor_date = disease_dicts['follow_up_start'][patient_key_from_ecg] + datetime.timedelta(
                    days=YEAR_DAYS * disease_dicts['follow_up_total'][patient_key_from_ecg],
                )
            else:
                has_disease = 1
                censor_date = disease_dicts['diagnosis_dates'][patient_key_from_ecg]

            if incidence_only and censor_date <= assess_date and has_disease:
                raise ValueError(f'{tm.name} only considers incident diagnoses')

            tensor[(i, 0) if dynamic else 0] = has_disease
            tensor[(i, 1) if dynamic else 1] = (censor_date - assess_date).days
        return tensor
    return _cox_tensor_from_file


def _survival_from_file(
    day_window: int, file_name: str, incidence_only: bool = False, patient_column: str = 'Mrn',
    follow_up_start_column: str = 'start_fu', follow_up_total_column: str = 'total_fu',
    diagnosis_column: str = 'first_stroke', delimiter: str = ',',
) -> Callable:
    """Build a tensor_from_file function for modeling survival curves of diagnoses given a TSV of patients and dates.

    The tensor_from_file function returned here should be used
    with SURVIVAL_CURVE TensorMaps to model survival curves of patients for a given disease.

    :param day_window: Total number of days of follow up the length of the survival curves to learn.
    :param file_name: CSV or TSV file with header of patient IDs (MRNs) dates of enrollment and dates of diagnosis
    :param incidence_only: Flag to skip patients whose diagnosis date is prior to acquisition date of input data
    :param patient_column: The header name of the column of patient ids
    :param follow_up_start_column: The header name of the column of enrollment dates
    :param follow_up_total_column: The header name of the column with total enrollment time (in years)
    :param diagnosis_column: The header name of the column of disease diagnosis dates
    :param delimiter: The delimiter separating columns of the TSV or CSV
    :return: The tensor_from_file function to provide to TensorMap constructors
    """
    error = None
    disease_dicts = defaultdict(dict)
    try:
        with open(file_name, 'r', encoding='utf-8') as f:
            reader = csv.reader(f, delimiter=delimiter)
            header = next(reader)
            follow_up_start_index = header.index(follow_up_start_column)
            follow_up_total_index = header.index(follow_up_total_column)
            patient_index = header.index(patient_column)
            date_index = header.index(diagnosis_column)
            for row in reader:
                try:
                    patient_key = int(row[patient_index])
                    disease_dicts['follow_up_start'][patient_key] = _loyalty_str2date(row[follow_up_start_index])
                    disease_dicts['follow_up_total'][patient_key] = float(row[follow_up_total_index])
                    if row[date_index] == '' or row[date_index] == 'NULL':
                        continue
                    disease_dicts['diagnosis_dates'][patient_key] = _loyalty_str2date(row[date_index])
                    if len(disease_dicts['follow_up_start']) % 2000 == 0:
                        logging.debug(f"Processed: {len(disease_dicts['follow_up_start'])} patient rows.")
                except ValueError as e:
                    logging.warning(f'val err {e}')
            logging.info(f"Done processing {diagnosis_column} Got {len(disease_dicts['follow_up_start'])} patient rows and {len(disease_dicts['diagnosis_dates'])} events.")
    except FileNotFoundError as e:
        error = e

    def tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        if error:
            raise error

        ecg_dates = _get_ecg_dates(tm, hd5)
        if len(ecg_dates) > 1:
            raise NotImplementedError('Survival curve models for multiple ECGs are not implemented.')
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        survival_then_censor = np.zeros(shape, dtype=np.float32)
        for ed, ecg_date in enumerate(ecg_dates):
            patient_key_from_ecg = _hd5_filename_to_mrn_int(hd5.filename)
            if patient_key_from_ecg not in disease_dicts['follow_up_start']:
                raise KeyError(f'{tm.name} mrn not in incidence csv')

            path = _make_hd5_path(tm, ecg_date, 'acquisitiondate')
            assess_date = _partners_str2date(decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype']))
            if assess_date < disease_dicts['follow_up_start'][patient_key_from_ecg]:
                raise ValueError(f'Assessed earlier than enrollment.')

            if patient_key_from_ecg not in disease_dicts['diagnosis_dates']:
                has_disease = 0
                censor_date = disease_dicts['follow_up_start'][patient_key_from_ecg] + datetime.timedelta(days=YEAR_DAYS*disease_dicts['follow_up_total'][patient_key_from_ecg])
            else:
                has_disease = 1
                censor_date = disease_dicts['diagnosis_dates'][patient_key_from_ecg]

            intervals = int(shape[1] if dynamic else shape[0] / 2)
            days_per_interval = day_window / intervals

            for i, day_delta in enumerate(np.arange(0, day_window, days_per_interval)):
                cur_date = assess_date + datetime.timedelta(days=day_delta)
                survival_then_censor[(ed, i) if dynamic else i] = float(cur_date < censor_date)
                survival_then_censor[(ed, intervals+i) if dynamic else intervals+i] = has_disease * float(censor_date <= cur_date < censor_date + datetime.timedelta(days=days_per_interval))
                if i == 0 and censor_date <= cur_date:  # Handle prevalent diseases
                    survival_then_censor[(ed, intervals) if dynamic else intervals] = has_disease
                    if has_disease and incidence_only:
                        raise ValueError(f'{tm.name} is skipping prevalent cases.')
            logging.debug(
                f"Got survival disease {has_disease}, censor: {censor_date}, assess {assess_date}, fu start {disease_dicts['follow_up_start'][patient_key_from_ecg]} "
                f"fu total {disease_dicts['follow_up_total'][patient_key_from_ecg]} tensor:{(survival_then_censor[ed] if dynamic else survival_then_censor)[:4]} mid tense: {(survival_then_censor[ed] if dynamic else survival_then_censor)[intervals:intervals+4]} ",
            )
        return survival_then_censor
    return tensor_from_file


def make_partners_diagnosis_maps(desired_map_name: str) -> Union[TensorMap, None]:
    diagnosis_columns = ['AF', 'CAD', 'DM', 'MI']
    if desired_map_name in diagnosis_columns:
        # Build survival curve TensorMaps
        return TensorMap(f'{desired_map_name.lower()}_event', Interpretation.SURVIVAL_CURVE, shape=(50,))


def make_wide_file_maps(desired_map_name: str) -> Union[TensorMap, None]:
    days_window = 1825

    if desired_map_name == 'sex_from_wide_csv':
        csv_tff = tensor_from_wide(WIDE_FILE, target='sex')
        return TensorMap(
            'sex_from_wide', Interpretation.CATEGORICAL, annotation_units=2, tensor_from_file=csv_tff,
            channel_map={'female': 0, 'male': 1},
        )
    elif desired_map_name == 'age_from_wide_csv':
        csv_tff = tensor_from_wide(WIDE_FILE, target='age')
        return TensorMap(
            'age_from_wide', Interpretation.CONTINUOUS, shape=(1,),
            tensor_from_file=csv_tff, channel_map={'age': 0},
            normalization={'mean': 63.35798891483556, 'std': 7.554638350423902},
        )
    elif desired_map_name == 'bmi_from_wide_csv':
        csv_tff = csv_tff = tensor_from_wide(WIDE_FILE, target='bmi')
        return TensorMap(
            'bmi_from_wide', Interpretation.CONTINUOUS, shape=(1,), channel_map={'bmi': 0},
            annotation_units=1, normalization={'mean': 27.3397, 'std': 4.77216}, tensor_from_file=csv_tff,
        )
    elif desired_map_name == 'ecg_2500_from_wide_csv':
        tff = tensor_from_wide(WIDE_FILE, target='ecg')
        return TensorMap(
            'ecg_rest_raw', shape=(2500, 12), path_prefix=PARTNERS_PREFIX, tensor_from_file=tff,
            cacheable=False, channel_map=ECG_REST_UKB_LEADS,
        )
    elif desired_map_name == 'ecg_5000_from_wide_csv':
        tff = tensor_from_wide(WIDE_FILE, target='ecg')
        return TensorMap(
            'ecg_rest_raw', shape=(5000, 12), path_prefix=PARTNERS_PREFIX, tensor_from_file=tff,
            cacheable=False, channel_map=ECG_REST_UKB_LEADS,
        )
    elif desired_map_name == 'time_to_hf_wide_csv':
        tff = tensor_from_wide(WIDE_FILE, target='time_to_event')
        return TensorMap('time_to_hf', Interpretation.TIME_TO_EVENT, tensor_from_file=tff)
    elif desired_map_name == 'survival_curve_hf_wide_csv':
        tff = tensor_from_wide(WIDE_FILE, target='survival_curve')
        return TensorMap('survival_curve_hf', Interpretation.SURVIVAL_CURVE, tensor_from_file=tff, shape=(50,), days_window=days_window)


def _date_from_dates(ecg_dates, target_date=None, earliest_date=None):
    if target_date:
        if target_date and earliest_date:
            incident_dates = [d for d in ecg_dates if earliest_date < datetime.datetime.strptime(d, PARTNERS_DATETIME_FORMAT) < target_date]
        else:
            incident_dates = [d for d in ecg_dates if datetime.datetime.strptime(d, PARTNERS_DATETIME_FORMAT) < target_date]
        if len(incident_dates) == 0:
            raise ValueError('No ECGs prior to target were found.')
        return np.random.choice(incident_dates)
    return np.random.choice(ecg_dates)


def _ecg_tensor_from_date(tm: TensorMap, hd5: h5py.File, ecg_date: str, population_normalize: int = None):
    tensor = np.zeros(tm.shape, dtype=np.float32)
    for cm in tm.channel_map:
        path = _make_hd5_path(tm, ecg_date, cm)
        voltage = decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype'])
        voltage = _resample_voltage(voltage, tm.shape[0])
        tensor[..., tm.channel_map[cm]] = voltage
    if population_normalize is not None:
        tensor /= population_normalize
    return tensor


def _to_float_or_none(s):
    try:
        return float(s)
    except ValueError:
        return None


def _days_to_years_float(s: str):
    try:
        return float(s.split(' ')[0]) / YEAR_DAYS
    except ValueError:
        return None


def _time_to_event_tensor_from_days(tm: TensorMap, has_disease: int, follow_up_days: int):
    tensor = np.zeros(tm.shape, dtype=np.float32)
    if follow_up_days > tm.days_window:
        has_disease = 0
        follow_up_days = tm.days_window + 1
    tensor[0] = has_disease
    tensor[1] = follow_up_days
    return tensor


def _survival_curve_tensor_from_dates(tm: TensorMap, has_disease: int, assessment_date: datetime.datetime, censor_date: datetime.datetime):
    intervals = tm.shape[0] // 2
    days_per_interval = tm.days_window / intervals
    survival_then_censor = np.zeros(tm.shape, dtype=np.float32)
    for i, day_delta in enumerate(np.arange(0, tm.days_window, days_per_interval)):
        cur_date = assessment_date + datetime.timedelta(days=day_delta)
        survival_then_censor[i] = float(cur_date < censor_date)
        survival_then_censor[intervals + i] = has_disease * float(censor_date <= cur_date < censor_date + datetime.timedelta(days=days_per_interval))
        if i == 0 and censor_date <= cur_date:  # Handle prevalent diseases
            survival_then_censor[intervals] = has_disease
    return survival_then_censor


def tensor_from_wide(
    file_name: str, patient_column: str = 'Mrn', age_column: str = 'age', bmi_column: str = 'bmi',
    sex_column: str = 'sex', hf_column: str = 'any_hf_age', start_column: str = 'start_fu',
    end_column: str = 'last_encounter', delimiter: str = '\t', population_normalize: int = 2000,
    target: str = 'ecg', skip_prevalent: bool = True,
) -> Callable:
    """Build a tensor_from_file function for ECGs in the legacy cohort.

    :param file_name: CSV or TSV file with header of patient IDs (MRNs) dates of enrollment and dates of diagnosis
    :param patient_column: The header name of the column of patient ids
    :param delimiter: The delimiter separating columns of the TSV or CSV
    :return: The tensor_from_file function to provide to TensorMap constructors
    """
    with open(file_name, 'r', encoding='utf-8') as f:
        reader = csv.reader(f, delimiter=delimiter)
        header = next(reader)
        hf_index = header.index(hf_column)
        age_index = header.index(age_column)
        bmi_index = header.index(bmi_column)
        sex_index = header.index(sex_column)
        end_index = header.index(end_column)
        start_index = header.index(start_column)
        patient_index = header.index(patient_column)
        patient_data = defaultdict(dict)
        for row in reader:
            try:
                patient_key = int(float(row[patient_index]))
                patient_data[patient_key] = {
                    'age': _days_to_years_float(row[age_index]),
                    'bmi': _to_float_or_none(row[bmi_index]),
                    'sex': row[sex_index],
                    'hf_age': _days_to_years_float(row[hf_index]),
                    'end_age': _days_to_years_float(row[end_index]),
                    'start_date': datetime.datetime.strptime(row[start_index], CARDIAC_SURGERY_DATE_FORMAT),
                }

            except ValueError as e:
                logging.debug(f'val err {e}')
        logging.info(f'Done processing. Got {len(patient_data)} patient rows.')

    def tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        mrn_int = _hd5_filename_to_mrn_int(hd5.filename)
        if mrn_int not in patient_data:
            raise KeyError(f'{tm.name} mrn not in csv.')
        if patient_data[mrn_int]['end_age'] is None or patient_data[mrn_int]['age'] is None:
            raise ValueError(f'{tm.name} could not find ages.')
        if patient_data[mrn_int]['end_age'] - patient_data[mrn_int]['age'] < 0:
            raise ValueError(f'{tm.name} has negative follow up time.')

        if patient_data[mrn_int]['hf_age'] is None:
            has_disease = 0
            follow_up_days = (patient_data[mrn_int]['end_age'] - patient_data[mrn_int]['age']) * YEAR_DAYS
        elif patient_data[mrn_int]['hf_age'] > patient_data[mrn_int]['age']:
            has_disease = 1
            follow_up_days = (patient_data[mrn_int]['hf_age'] - patient_data[mrn_int]['age']) * YEAR_DAYS
        elif skip_prevalent and patient_data[mrn_int]['age'] > patient_data[mrn_int]['hf_age']:
            raise ValueError(f'{tm.name} skips prevalent cases.')
        else:
            has_disease = 1
            follow_up_days = (patient_data[mrn_int]['hf_age'] - patient_data[mrn_int]['age']) * YEAR_DAYS

        if target == 'time_to_event':
            tensor = _time_to_event_tensor_from_days(tm, has_disease, follow_up_days)
            logging.debug(f'Returning {tensor} for {patient_data[mrn_int]} key {mrn_int}')
            return tensor
        elif target == 'survival_curve':
            end_date = patient_data[mrn_int]['start_date'] + datetime.timedelta(days=follow_up_days)
            tensor = _survival_curve_tensor_from_dates(tm, has_disease, patient_data[mrn_int]['start_date'], end_date)
            logging.debug(
                f"Got survival disease {has_disease}, censor: {end_date}, assess {patient_data[mrn_int]['start_date']}, age {patient_data[mrn_int]['age']} "
                f"end age: {patient_data[mrn_int]['end_age']} hf age: {patient_data[mrn_int]['hf_age']} "
                f"fu total {follow_up_days/YEAR_DAYS} tensor:{tensor[:4]} mid tense: {tensor[tm.shape[0] // 2:(tm.shape[0] // 2)+4]} ",
            )
            return tensor
        elif target == 'ecg':
            ecg_dates = list(hd5[tm.path_prefix])
            earliest = patient_data[mrn_int]['start_date'] - datetime.timedelta(days=3*YEAR_DAYS)
            ecg_date_key = _date_from_dates(ecg_dates, patient_data[mrn_int]['start_date'], earliest)
            return _ecg_tensor_from_date(tm, hd5, ecg_date_key, population_normalize)
        elif target in ['age', 'bmi']:
            tensor = np.zeros(tm.shape, dtype=np.float32)
            if patient_data[mrn_int][target] is None:
                raise ValueError(f'Missing target value {target}')
            tensor[0] = patient_data[mrn_int][target]
            return tensor
        elif target == 'sex':
            tensor = np.zeros(tm.shape, dtype=np.float32)
            if patient_data[mrn_int][target].lower() == 'female':
                tensor[0] = 1.0
            elif patient_data[mrn_int][target].lower() == 'male':
                tensor[1] = 1.0
            logging.debug(f'Returning {tensor} for {patient_data[mrn_int][target]} key {mrn_int}')
            return tensor
        else:
            raise ValueError(f'{tm.name} has no way to handle target {target}')
    return tensor_from_file


def build_cardiac_surgery_dict(
    filename: str = CARDIAC_SURGERY_OUTCOMES_CSV,
    patient_column: str = 'medrecn',
    date_column: str = 'surgdt',
    additional_columns: List[str] = [],
) -> Dict[int, Dict[str, Union[int, str]]]:
    keys = [date_column] + additional_columns
    cardiac_surgery_dict = {}
    df = pd.read_csv(
        filename,
        low_memory=False,
        usecols=[patient_column]+keys,
    ).sort_values(by=[patient_column, date_column])
    # sort dataframe such that newest surgery per patient appears later and is used in lookup table
    for row in df.itertuples():
        patient_key = getattr(row, patient_column)
        cardiac_surgery_dict[patient_key] = {key: getattr(row, key) for key in keys}
    return cardiac_surgery_dict


def build_date_interval_lookup(
    cardiac_surgery_dict: Dict[int, Dict[str, Union[int, str]]],
    start_column: str = 'surgdt',
    start_offset: int = -30,
    end_column: str = 'surgdt',
    end_offset: int = 0,
) -> Dict[int, Tuple[str, str]]:
    date_interval_lookup = {}
    for mrn in cardiac_surgery_dict:
        start_date = (_cardiac_surgery_str2date(cardiac_surgery_dict[mrn][start_column], PARTNERS_DATETIME_FORMAT.replace('T', ' ')) + datetime.timedelta(days=start_offset)).strftime(PARTNERS_DATETIME_FORMAT)
        end_date = (_cardiac_surgery_str2date(cardiac_surgery_dict[mrn][end_column], PARTNERS_DATETIME_FORMAT.replace('T', ' ')) + datetime.timedelta(days=end_offset)).strftime(PARTNERS_DATETIME_FORMAT)
        date_interval_lookup[mrn] = (start_date, end_date)
    return date_interval_lookup


def make_cardiac_surgery_outcome_tensor_from_file(
    cardiac_surgery_dict: Dict[int, Dict[str, Union[int, str]]],
    outcome_column: str,
) -> Callable:
    def tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}):
        mrn = _hd5_filename_to_mrn_int(hd5.filename)
        tensor = np.zeros(tm.shape, dtype=np.float32)
        outcome = cardiac_surgery_dict[mrn][outcome_column]

        if type(outcome) is float and not outcome.is_integer():
            raise ValueError(f'Cardiac Surgery categorical outcome {tm.name} ({outcome_column}) got non-discrete value: {outcome}')

        # ensure binary outcome
        if outcome != 0 and outcome != 1:
            raise ValueError(f'Cardiac Surgery categorical outcome {tm.name} ({outcome_column}) got non-binary value: {outcome}')

        tensor[outcome] = 1
        return tensor
    return tensor_from_file


def build_cardiac_surgery_tensor_maps(
    needed_name: str,
) -> TensorMap:
    outcome2column = {
        "sts_death": "mtopd",
        "sts_stroke": "cnstrokp",
        "sts_renal_failure": "crenfail",
        "sts_prolonged_ventilation": "cpvntlng",
        "sts_dsw_infection": "deepsterninf",
        "sts_reoperation": "reop",
        "sts_any_morbidity": "anymorbidity",
        "sts_long_stay": "llos",
    }

    cardiac_surgery_dict = None
    date_interval_lookup = None
    if needed_name in outcome2column:
        if cardiac_surgery_dict is None:
            cardiac_surgery_dict = build_cardiac_surgery_dict(
                additional_columns=[outcome2column[needed_name]],
            )
            channel_map = _outcome_channels(needed_name)
            sts_tmap = TensorMap(
                needed_name,
                Interpretation.CATEGORICAL,
                path_prefix=PARTNERS_PREFIX,
                tensor_from_file=make_cardiac_surgery_outcome_tensor_from_file(
                    cardiac_surgery_dict, outcome2column[needed_name],
                ),
                channel_map=channel_map,
                validator=validator_not_all_zero,
            )
        else:
            if needed_name.endswith('_sts'):
                base_name = needed_name.split('_sts')[0]
                tmap_map = build_partners_time_series_tensor_maps([base_name])

                if cardiac_surgery_dict is None:
                    cardiac_surgery_dict = build_cardiac_surgery_dict(
                        additional_columns=[outcome2column[needed_name]],
                    )
                if date_interval_lookup is None:
                    date_interval_lookup = build_date_interval_lookup(cardiac_surgery_dict)
                sts_tmap = copy.deepcopy(tmap_map[base_name])
                sts_tmap.name = needed_name
                sts_tmap.time_series_lookup = date_interval_lookup

    return sts_tmap


def _get_measurement_matrix_entry(matrix: np.ndarray, key_idx: int, lead_idx: Union[int, List[int]] = None):
    # First 18 words of measurement matrix are for global measurements, then each lead has 53*2 words
    lead_start = 18
    lead_words = 53 * 2
    # In the measurement matrix, NA values are encoded as 2^15 - 1. Introducing here a threshold
    nan_threshold = 32765
    if lead_idx is None:
        idx = key_idx
    elif isinstance(lead_idx, list):
        matrix_values = []
        for lead_index in lead_idx:
            idx = lead_start + lead_index * lead_words + (key_idx-1)*2+1
            if matrix[idx] > nan_threshold:
                matrix_values.append(np.nan)
            else:
                matrix_values.append(matrix[idx])
        return max(matrix_values)
    else:
        idx = lead_start + lead_idx * lead_words + (key_idx-1)*2+1
    value = np.nan if matrix[idx] > nan_threshold else matrix[idx]
    return value


def make_measurement_matrix_from_file(key_idx: int, lead_idx: Union[int, List[int]] = None):
    def measurement_matrix_from_file(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        tensor = np.zeros(shape, dtype=float)
        for i, ecg_date in enumerate(ecg_dates):
            path = _make_hd5_path(tm, ecg_date, 'measurementmatrix')
            matrix = decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype'])
            tensor[i] = _get_measurement_matrix_entry(matrix, key_idx, lead_idx)
        return tensor
    return measurement_matrix_from_file


def make_mgb_ecg_measurement_matrix_global_tensor_maps(needed_name: str):
    # Measurement matrix TMAPS -- indices from MUSE XML dev manual, page 49 and following
    measurement_matrix_global_measures = {
        'pon': 1,       # P-wave onset in median beat (in samples)
        'poff': 2,      # P-wave offset in median beat
        'qon': 3,       # Q-Onset in median beat
        'qoff': 4,      # Q-Offset in median beat
        'ton': 5,       # T-Onset in median beat
        'toff': 6,      # T-Offset in median beat
        'nqrs': 7,      # Number of QRS Complexes
        'qrsdur': 8,    # QRS Duration
        'qt': 9,        # QT Interval
        'qtc': 10,      # QT Corrected
        'print': 11,    # PR Interval
        'vrate': 12,    # Ventricular Rate
        'avgrr': 13,    # Average R-R Interval
    }
    for measure, measure_idx in measurement_matrix_global_measures.items():
        if f'partners_ecg_measurement_matrix_{measure}' == needed_name:
            return TensorMap(
                f'partners_ecg_measurement_matrix_{measure}',
                interpretation=Interpretation.CONTINUOUS,
                shape=(None, 1),
                path_prefix=PARTNERS_PREFIX,
                loss='log_cosh',
                time_series_limit=0,
                tensor_from_file=make_measurement_matrix_from_file(measure_idx),
            )


measurement_matrix_leads = {
    'I': 0, 'II': 1, 'V1': 2, 'V2': 3, 'V3': 4, 'V4':5, 'V5': 6, 'V6': 7, 'III': 8, 'aVR': 9, 'aVL': 10, 'aVF': 11,
}
measurement_matrix_lead_measures = {
    'pona': 1,      # P Wave amplitude at P-onset
    'pamp': 2,      # P wave amplitude
    'pdur': 3,      # P wave duration
    'bmpar': 4,     # P wave area
    'bmpi': 5,      # P wave intrinsicoid (time from P onset to peak of P)
    'ppamp': 6,     # P Prime amplitude
    'ppdur': 7,     # P Prime duration
    'bmppar': 8,    # P Prime area
    'bmppi': 9,     # P Prime intrinsicoid (time from P onset to peak of P')
    'qamp': 10,     # Q wave amplitude
    'qdur': 11,     # Q wave duration
    'bmqar': 12,    # Q wave area
    'bmqi': 13,     # Q intrinsicoid (time from Q onset to peak of Q)
    'ramp': 14,     # R amplitude
    'rdur': 15,     # R duration
    'bmrar': 16,    # R wave area
    'bmri': 17,     # R intrinsicoid (time from R onset to peak of R)
    'samp': 18,     # S amplitude
    'sdur': 19,     # S duration
    'bmsar': 20,    # S wave area
    'bmsi': 21,     # S intrinsicoid (time from Q onset to peak of S)
    'rpamp': 22,    # R Prime amplitude
    'rpdur': 23,    # R Prime duration
    'bmrpar': 24,   # R Prime wave area
    'bmrpi': 25,    # R Prime intrinsicoid (time from Q onset to peak of R Prime)
    'spamp': 26,    # S Prime Amplitude
    'spdur': 27,    # S Prime Duration
    'bmspar': 28,   # S Prime wave area
    'bmspi': 29,    # S intriniscoid (time from Q onset to peak of S prime)
    'stj': 30,      # STJ point, End of QRS Point Amplitude
    'stm': 31,      # STM point, Middle of the ST Segment Amplitude
    'ste': 32,      # STE point, End of ST Segment Amplitude
    'mxsta': 33,    # Maximum of STJ, STM, STE Amplitudes
    'mnsta': 34,    # Minimum of STJ and STM Amplitudes
    'spta': 35,     # Special T-Wave amplitude
    'qrsa': 36,     # Total QRS area
    'qrsdef': 37,   # QRS Deflection
    'maxra': 38,    # Maximum R Amplitude (R or R Prime)
    'maxsa': 39,    # Maximum S Amplitude (S or S Prime)
    'tamp': 40,     # T amplitude
    'tdur': 41,     # T duration
    'bmtar': 42,    # T wave area
    'bmti': 43,     # T intriniscoid (time from STE to peak of T)
    'tpamp': 44,    # T Prime amplitude
    'tpdur': 45,    # T Prime duration
    'bmtpar': 46,   # T Prime area
    'bmtpi': 47,    # T Prime intriniscoid (time from STE to peak of T)
    'tend': 48,     # T Amplitude at T offset
    'parea': 49,    # P wave area, includes P and P Prime
    'qrsar': 50,    # QRS area
    'tarea': 51,    # T wave area, include T and T Prime
    'qrsint': 52,    # QRS intriniscoid (see following)
}


def make_mgb_ecg_measurement_matrix_lead_tensor_maps(needed_name: str):
    for lead, lead_idx in measurement_matrix_leads.items():
        for measure, measure_idx in measurement_matrix_lead_measures.items():
            if f'partners_ecg_measurement_matrix_{lead}_{measure}' == needed_name:
                return TensorMap(
                    f'partners_ecg_measurement_matrix_{lead}_{measure}',
                    interpretation=Interpretation.CONTINUOUS,
                    shape=(None, 1),
                    path_prefix=PARTNERS_PREFIX,
                    loss='log_cosh',
                    time_series_limit=0,
                    tensor_from_file=make_measurement_matrix_from_file(measure_idx, lead_idx=lead_idx),
                )

    # Max values across leads
    for measure, measure_idx in measurement_matrix_lead_measures.items():
        if f'partners_ecg_measurement_matrix_max_{measure}' == needed_name:
            return TensorMap(
                f'partners_ecg_measurement_matrix_max_{measure}',
                interpretation=Interpretation.CONTINUOUS,
                shape=(None, 1),
                path_prefix=PARTNERS_PREFIX,
                loss='log_cosh',
                time_series_limit=0,
                tensor_from_file=make_measurement_matrix_from_file(measure_idx, lead_idx=list(measurement_matrix_leads.values())),
            )


def make_mgb_ecg_lvh_tensormaps(needed_name: str):
    def ecg_lvh_from_file(tm: TensorMap, hd5: h5py.File, dependents={}):
        # Lead order seems constant and standard throughout, but we could eventually tensorize it from XML
        avl_min = 1100.0
        sl_min = 3500.0
        cornell_female_min = 2000.0
        cornell_male_min = 2800.0
        sleads = ['V1', 'V3']
        rleads = ['aVL', 'V5', 'V6']
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        tensor = np.zeros(shape, dtype=float)

        for i, ecg_date in enumerate(ecg_dates):
            path = _make_hd5_path(tm, ecg_date, 'measurementmatrix')
            matrix = decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype'])
            criteria_sleads = {lead: _get_measurement_matrix_entry(matrix, measurement_matrix_lead_measures['samp'], measurement_matrix_leads[lead]) for lead in sleads}
            criteria_rleads = {lead: _get_measurement_matrix_entry(matrix, measurement_matrix_lead_measures['ramp'], measurement_matrix_leads[lead]) for lead in rleads}
            if 'avl_lvh' in tm.name:
                is_lvh = np.nan if np.isnan(criteria_rleads['aVL']) else criteria_rleads['aVL'] > avl_min
            elif 'sokolow_lyon_lvh' in tm.name:
                criteria_sum = criteria_sleads['V1'] + np.maximum(criteria_rleads['V5'], criteria_rleads['V6'])
                is_lvh = np.nan if np.isnan(criteria_sum) else criteria_sum > sl_min
            elif 'cornell_lvh' in tm.name:
                is_lvh = criteria_rleads['aVL'] + criteria_sleads['V3']
                sex_path = _make_hd5_path(tm, ecg_date, 'gender')
                is_female = 'female' in decompress_data(data_compressed=hd5[sex_path][()], dtype='str').lower()
                if is_female:
                    is_lvh = np.nan if np.isnan(is_lvh) else is_lvh > cornell_female_min
                else:
                    is_lvh = np.nan if np.isnan(is_lvh) else is_lvh > cornell_male_min
            else:
                raise ValueError(f'{tm.name} criterion for LVH is not accounted for')
            # Following convention from categorical TMAPS, positive has cmap index 1
            if np.isnan(is_lvh):
                if dynamic:
                    tensor[i, :] = np.nan
                else:
                    tensor[:] = np.nan
            else:
                index = 1 if is_lvh else 0
                slices = (i, index) if dynamic else (index,)
                tensor[slices] = 1.0
        return tensor

    for criterion in ['avl_lvh', 'sokolow_lyon_lvh', 'cornell_lvh']:
        if f'partners_ecg_{criterion}' == needed_name:
            return TensorMap(
                f'partners_ecg_{criterion}',
                interpretation=Interpretation.CATEGORICAL,
                path_prefix=PARTNERS_PREFIX,
                tensor_from_file=ecg_lvh_from_file,
                channel_map={f'no_{criterion}': 0, criterion: 1},
                shape=(None, 2),
                time_series_limit=0,
            )



# Tensor maps  for the CelebA Dataset
# https://mmlab.ie.cuhk.edu.hk/projects/CelebA.html
import os
import sys
import gzip
import pickle
import logging
from typing import Dict

import h5py
import numpy as np

from skimage.transform import resize
from ml4h.normalizer import ZeroMeanStd1
from ml4h.TensorMap import TensorMap, Interpretation

def image_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    return np.array(hd5[tm.name][:tm.shape[0], :tm.shape[1], :tm.shape[2]], dtype=np.float32)


mnist_image = TensorMap('mnist_image', shape=(28, 28, 1), tensor_from_file=image_from_hd5)
celeba_image = TensorMap(
    'celeba_image', shape=(218, 178, 3), normalization=ZeroMeanStd1(),
    tensor_from_file=image_from_hd5,
)
celeba_image_208 = TensorMap(
    'celeba_image', shape=(208, 168, 3), normalization=ZeroMeanStd1(),
    tensor_from_file=image_from_hd5,
)
celeba_image_208_raw = TensorMap('celeba_image', shape=(208, 168, 3), tensor_from_file=image_from_hd5)


def downsampled_image_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    dependents[tm.dependent_map] = np.array(hd5[tm.dependent_map.name][:tm.dependent_map.shape[0], :tm.dependent_map.shape[1], :tm.dependent_map.shape[2]], dtype=np.float32)
    dependents[tm.dependent_map] -= dependents[tm.dependent_map].mean()
    dependents[tm.dependent_map] /= (1e-8+ dependents[tm.dependent_map].std())
    resized = resize(dependents[tm.dependent_map], (tm.shape[0], tm.shape[1]))
    return resized


celeba_image_208_downsample_2x = TensorMap(
    'celeba_image_208_downsample_2x', shape=(104, 84, 3),
    tensor_from_file=downsampled_image_from_hd5,
    dependent_map=celeba_image_208_raw,
)

celeba_image_208_downsample_4x = TensorMap(
    'celeba_image_208_downsample_4x', shape=(52, 42, 3),
    tensor_from_file=downsampled_image_from_hd5,
    dependent_map=celeba_image_208_raw,
)


def landmark_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    landmark = np.zeros(tm.shape, dtype=np.float32)
    for channel in tm.channel_map:
        landmark[tm.channel_map[channel]] = hd5[channel][0]
    return landmark


eye_channels = {'lefteye_x': 0, 'lefteye_y': 1, 'righteye_x': 2, 'righteye_y': 3}
celeba_eyes = TensorMap(
    'celeba_eyes', shape=(len(eye_channels),), channel_map=eye_channels,
    tensor_from_file=landmark_from_hd5,
)


def celeba_label(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    if tm.name in hd5:
        return np.array([1, 0])
    else:
        return np.array([0, 1])


celeba_arched_eyebrows = TensorMap('arched_eyebrows', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_arched_eyebrows': 0, 'arched_eyebrows': 1})
celeba_attractive = TensorMap('attractive', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_attractive': 0, 'attractive': 1})
celeba_bags_under_eyes = TensorMap('bags_under_eyes', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_bags_under_eyes': 0, 'bags_under_eyes': 1})
celeba_bald = TensorMap('bald', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_bald': 0, 'bald': 1})
celeba_bangs = TensorMap('bangs', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_bangs': 0, 'bangs': 1})
celeba_big_lips = TensorMap('big_lips', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_big_lips': 0, 'big_lips': 1})
celeba_big_nose = TensorMap('big_nose', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_big_nose': 0, 'big_nose': 1})
celeba_black_hair = TensorMap('black_hair', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_black_hair': 0, 'black_hair': 1})
celeba_blond_hair = TensorMap('blond_hair', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_blond_hair': 0, 'blond_hair': 1})
celeba_blurry = TensorMap('blurry', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_blurry': 0, 'blurry': 1})
celeba_brown_hair = TensorMap('brown_hair', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_brown_hair': 0, 'brown_hair': 1})
celeba_bushy_eyebrows = TensorMap('bushy_eyebrows', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_bushy_eyebrows': 0, 'bushy_eyebrows': 1})
celeba_chubby = TensorMap('chubby', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_chubby': 0, 'chubby': 1})
celeba_double_chin = TensorMap('double_chin', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_double_chin': 0, 'double_chin': 1})
celeba_eyeglasses = TensorMap('eyeglasses', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_eyeglasses': 0, 'eyeglasses': 1})
celeba_goatee = TensorMap('goatee', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_goatee': 0, 'goatee': 1})
celeba_gray_hair = TensorMap('gray_hair', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_gray_hair': 0, 'gray_hair': 1})
celeba_heavy_makeup = TensorMap('heavy_makeup', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_heavy_makeup': 0, 'heavy_makeup': 1})
celeba_high_cheekbones = TensorMap('high_cheekbones', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_high_cheekbones': 0, 'high_cheekbones': 1})
celeba_male = TensorMap('male', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_male': 0, 'male': 1})
celeba_mouth_slightly_open = TensorMap('mouth_slightly_open', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_mouth_slightly_open': 0, 'mouth_slightly_open': 1})
celeba_mustache = TensorMap('mustache', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_mustache': 0, 'mustache': 1})
celeba_narrow_eyes = TensorMap('narrow_eyes', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_narrow_eyes': 0, 'narrow_eyes': 1})
celeba_no_beard = TensorMap('no_beard', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_no_beard': 0, 'no_beard': 1})
celeba_oval_face = TensorMap('oval_face', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_oval_face': 0, 'oval_face': 1})
celeba_pale_skin = TensorMap('pale_skin', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_pale_skin': 0, 'pale_skin': 1})
celeba_pointy_nose = TensorMap('pointy_nose', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_pointy_nose': 0, 'pointy_nose': 1})
celeba_receding_hairline = TensorMap('receding_hairline', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_receding_hairline': 0, 'receding_hairline': 1})
celeba_rosy_cheeks = TensorMap('rosy_cheeks', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_rosy_cheeks': 0, 'rosy_cheeks': 1})
celeba_sideburns = TensorMap('sideburns', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_sideburns': 0, 'sideburns': 1})
celeba_smiling = TensorMap('smiling', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_smiling': 0, 'smiling': 1})
celeba_straight_hair = TensorMap('straight_hair', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_straight_hair': 0, 'straight_hair': 1})
celeba_wavy_hair = TensorMap('wavy_hair', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_wavy_hair': 0, 'wavy_hair': 1})
celeba_wearing_earrings = TensorMap('wearing_earrings', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_wearing_earrings': 0, 'wearing_earrings': 1})
celeba_wearing_hat = TensorMap('wearing_hat', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_wearing_hat': 0, 'wearing_hat': 1})
celeba_wearing_lipstick = TensorMap('wearing_lipstick', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_wearing_lipstick': 0, 'wearing_lipstick': 1})
celeba_wearing_necklace = TensorMap('wearing_necklace', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_wearing_necklace': 0, 'wearing_necklace': 1})
celeba_wearing_necktie = TensorMap('wearing_necktie', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_wearing_necktie': 0, 'wearing_necktie': 1})
celeba_young = TensorMap('young', Interpretation.CATEGORICAL, tensor_from_file=celeba_label, channel_map={'no_young': 0, 'young': 1})



"""Methods for batch annotations of images stored as 3D tensors, such as MRIs, from within notebooks."""

import json
import os
import socket
import tempfile
from typing import Any, Dict, List, Tuple, Type, Union

from IPython.display import display
import numpy as np
import pandas as pd
import h5py
from ipyannotations.images import PolygonAnnotator
from ipyannotations.images.annotator import Annotator
import ipywidgets as widgets
from ml4h.visualization_tools.hd5_mri_plots import MRI_TMAPS
from ml4h.visualization_tools.annotation_storage import AnnotationStorage
from ml4h.visualization_tools.annotation_storage import TransientAnnotationStorage
from PIL import Image
import tensorflow as tf


class BatchImageAnnotator():
  """Annotate batches of images with shapes drawn over regions of interest."""

  SUBMIT_BUTTON_DESCRIPTION = 'Submit annotations, goto next sample'
  EXPECTED_COLUMN_NAMES = ['sample_id', 'tmap_name', 'instance_number', 'folder']
  DEFAULT_ANNOTATION_CLASSNAME = 'region_of_interest'
  CSS = '''
      <style>
      table, th, td {
        border: 2px solid black;
        border-collapse: collapse;
      }
      th, td {
        padding-left: 5px;
      }
      </style>
      '''

  def __init__(
      self, samples: pd.DataFrame, annotation_categories: List[str] = None,
      zoom: float = 1.5, annotation_storage: AnnotationStorage = TransientAnnotationStorage(),
      annotator: Type[Annotator] = PolygonAnnotator,
  ):
    """Initializes an instance of BatchImageAnnotator.

    Args:
      samples: A dataframe of samples to annotate. Columns must include those
               in BatchImageAnnotator.EXPECTED_COLUMN_NAMES.
      annotation_categories: A list of one or more strings to serve as tags for the annotations.
      zoom: Desired zoom level for the image. Defaults to 1.5.
      annotation_storage: An instance of AnnotationStorage. This faciltates the use of a user-provided
          strategy for the storage and processing of annotations. Defaults to  TransientAnnotationStorage.
      annotator: An instance of Annotator from package ipyannotations. Concrete classes include
        ipyannotations.BoxAnnotator, ipyannotations.PointAnnotator, and ipyannotations.PolygonAnnotator.
        Defaults to  ipyannotations.PolygonAnnotator.

    Raises:
      ValueError: The provided dataframe does not contain the expected columns.
    """
    if not set(self.EXPECTED_COLUMN_NAMES).issubset(samples.columns):
      raise ValueError(f'samples Dataframe must contain columns {self.EXPECTED_COLUMN_NAMES}')
    self.samples = samples
    self.current_sample = 0
    # TODO(deflaux) remove this after https://github.com/janfreyberg/ipyannotations/issues/11
    self.zoom = zoom
    self.annotation_storage = annotation_storage
    if annotation_categories is None:
      annotation_categories = [self.DEFAULT_ANNOTATION_CLASSNAME]

    self.annotation_widget = annotator(
        options=annotation_categories,
        canvas_size=(900, 280 * self.zoom),
    )
    self.annotation_widget.on_submit(self._store_annotations)
    self.annotation_widget.submit_button.description = self.SUBMIT_BUTTON_DESCRIPTION
    self.annotation_widget.submit_button.layout = widgets.Layout(width='300px')

    # Restructure the use instructions from the pydoc into a form that displays well as HTML.
    self.use_instructions = (
        '<ul><li>' +
        annotator.__doc__[0:annotator.__doc__.find('Parameters')].strip().replace('\n\n', '</li><li>') +
        '</li></ul>'
    )

    self.title_widget = widgets.HTML('')
    self.results_widget = widgets.HTML('')

  def _store_annotations(self, data: Dict[Any, Any]) -> None:
    """Transfer widget state to the annotation storage and advance to the next sample."""
    if self.current_sample >= self.samples.shape[0]:
      self.results_widget.value = '<h1>Annotation batch complete!</h1>Thank you for making the model better.'
      return

    # Convert canvas coordinates to tensor coordinates.
    image_canvas_position = self.annotation_widget.canvas.image_extent
    x_offset, y_offset, _, _ = image_canvas_position
    annotations = []
    for item in data:
      annotation: Dict[str, Union[List[Tuple[int, int]], Tuple[int, int], Tuple[int, int, int, int], str]] = {}
      annotations.append(annotation)
      for key in item.keys():
        if key == 'points':  # Polygons from PolygonAnnotator
          annotation[key] = [(
              int((p[0] - x_offset) / self.zoom),
              int((p[1] - y_offset) / self.zoom),
          ) for p in item[key]]
        elif key == 'coordinates':  # Points from PointAnnotator
          annotation[key] = (
              int((item[key][0] - x_offset) / self.zoom),
              int((item[key][1] - y_offset) / self.zoom),
          )
        elif key == 'xyxy':  # Rectangles from BoxAnnotator
          annotation[key] = (
              int((item[key][0] - x_offset) / self.zoom),
              int((item[key][1] - y_offset) / self.zoom),
              int((item[key][2] - x_offset) / self.zoom),
              int((item[key][3] - y_offset) / self.zoom),
          )
        else:
          # Pass all other values through unchanged.
          annotation[key] = item[key]

    # Store the annotation using the provided annotation storage strategy.
    self.annotation_storage.submit_annotation(
        sample_id=self.samples.loc[self.current_sample, 'sample_id'],
        annotator=os.getenv('OWNER_EMAIL') if os.getenv('OWNER_EMAIL') else socket.gethostname(),
        key=self.samples.loc[self.current_sample, 'tmap_name'],
        value_numeric=self.samples.loc[self.current_sample, 'instance_number'],
        value_string=self.samples.loc[self.current_sample, 'folder'],
        comment=json.dumps(annotations),
    )

    # Display this annotation at the bottom of the widget.
    results = f'''
        <hr>
        <h2>Prior sample's submitted annotations</h2>
        The <b>{self.SUBMIT_BUTTON_DESCRIPTION}</b> button is both printing out the annotations below and storing the annotations
        via strategy {self.annotation_storage.__class__.__name__}.<br>
        Details: <i>{self.annotation_storage.describe()}</i>
        <h3>sample info</h3>
        {self._format_info_for_current_sample()}
        <h3>canvas coordinates</h3>
        image extent {image_canvas_position}
        {[f'<pre>{json.dumps(x)}</pre>' for x in data]}
        <h3>source tensor coordinates</h3>
        {[f'<pre>{json.dumps(x)}</pre>' for x in annotations]}
        <hr>
      '''
    self.results_widget.value = results

    # Advance to the next sample.
    self.current_sample += 1
    self._annotate_image_for_current_sample()

  def _format_info_for_current_sample(self) -> str:
    """Convert information about the current sample to an HTML table for display within the widget."""
    headings = ' '.join([f'<th>{c}</th>' for c in self.EXPECTED_COLUMN_NAMES] + ['<th>TMAP shape</th>'])
    values = ' '.join([f'<td>{self.samples.loc[self.current_sample, c]}</td>' for c in self.EXPECTED_COLUMN_NAMES]
                      + [f'<td>{MRI_TMAPS[self.samples.loc[self.current_sample, "tmap_name"]].shape}</td>'])
    return f'''
        <table style="width:100%">
        <tr>{headings}</tr>
        <tr>{values}</tr>
        </table>
        '''

  def _annotate_image_for_current_sample(self) -> None:
    """Retrieve the data for the current sample and display its image in the annotation widget.

    If all samples have been processed, display the completion message.
    """
    if self.current_sample >= self.samples.shape[0]:
      self.annotation_widget.canvas.clear()
      self.title_widget.value = '<h1>Annotation batch complete!</h1>Thank you for making the model better.'
      return

    sample_id = self.samples.loc[self.current_sample, 'sample_id']
    tmap_name = self.samples.loc[self.current_sample, 'tmap_name']
    instance_number = self.samples.loc[self.current_sample, 'instance_number']
    folder = self.samples.loc[self.current_sample, 'folder']

    with tempfile.TemporaryDirectory() as tmpdirname:
      sample_hd5 = str(sample_id) + '.hd5'
      local_path = os.path.join(tmpdirname, sample_hd5)
      try:
        tf.io.gfile.copy(src=os.path.join(folder, sample_hd5), dst=local_path)
        hd5 = h5py.File(local_path, mode='r')
      except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
        self.annotation_widget.canvas.clear()
        self.title_widget.value = f'''
            <div class="alert alert-block alert-danger">
            <H2>Warning: MRI HD5 file not available for sample {sample_id} in folder {folder}</h2>
            Use the <kbd>folder</kbd> parameter to read HD5s from a different local directory or Cloud Storage bucket.
            <hr><p><pre>{e.message}</pre></p>
            </div>
            '''
        return

    tensor = MRI_TMAPS[tmap_name].tensor_from_file(MRI_TMAPS[tmap_name], hd5)
    tensor_instance = tensor[:, :, instance_number]
    if self.zoom > 1.0:
      # TODO(deflaux) remove this after https://github.com/janfreyberg/ipyannotations/issues/11
      img = Image.fromarray(tensor_instance)
      zoomed_img = img.resize([int(self.zoom * s) for s in img.size], Image.LANCZOS)
      tensor_instance = np.asarray(zoomed_img)

    self.annotation_widget.display(tensor_instance)
    self.title_widget.value = f'''
        {self.CSS}
        <div class='alert alert-block alert-info'>
        <h1>Batch annotation of {self.samples.shape[0]} samples</h1>
        {self.use_instructions}
        </div>
        <h2>Current sample</h2>
        {self._format_info_for_current_sample()}
        <hr>
    '''

  def annotate_images(self) -> None:
    """Begin the batch annotation task by displaying the annotation widget populated with the first sample.

    The submit button is used to proceed to the next sample until all samples have been processed.
    """
    self._annotate_image_for_current_sample()
    display(widgets.VBox([self.title_widget, self.annotation_widget, self.results_widget]))

  def view_recent_submissions(self, count: int = 10) -> pd.DataFrame:
    """View a dataframe of up to [count] most recent submissions.

    Args:
      count: The number of the most recent submissions to return.

    Returns:
      A dataframe of the most recent annotations.
    """
    return self.annotation_storage.view_recent_submissions(count=count)



"""Methods for Altair-based interactive plots for ECG data for use within notebooks."""

import os
import tempfile
from typing import Optional, Union

from IPython.display import HTML
import altair as alt  # Interactive data visualization for plots.
from ml4h.TensorMap import TensorMap
from ml4h.visualization_tools.ecg_reshape import DEFAULT_RESTING_ECG_SIGNAL_TMAP
from ml4h.visualization_tools.ecg_reshape import reshape_exercise_ecg_to_tidy
from ml4h.visualization_tools.ecg_reshape import reshape_resting_ecg_to_tidy

# Configure for large-data altair plotting.
# https://altair-viz.github.io/user_guide/faq.html#why-does-altair-lead-to-such-extremely-large-notebooks
alt.data_transformers.disable_max_rows()

# Define this at the global level so that the tempfile exists for the duration of the Jupyter kernel.
RESTING_ECG_DATA_FILE = tempfile.NamedTemporaryFile(
    dir=os.getcwd(),
    prefix='tidy_resting_ecg_',
    suffix='.json',
)
EXERCISE_ECG_TREND_DATA_FILE = tempfile.NamedTemporaryFile(
    dir=os.getcwd(),
    prefix='tidy_exercise_ecg_trend_',
    suffix='.json',
)
EXERCISE_ECG_SIGNAL_DATA_FILE = tempfile.NamedTemporaryFile(
    dir=os.getcwd(),
    prefix='tidy_exercise_ecg_signal_',
    suffix='.json',
)


def resting_ecg_interactive_plot(
    sample_id: Union[int, str], folder: Optional[str] = None,
    tmap: TensorMap = DEFAULT_RESTING_ECG_SIGNAL_TMAP,
) -> Union[HTML, alt.Chart]:
  """Wrangle resting ECG data to tidy and present it as an interactive plot.

  Args:
    sample_id: The id of the ECG sample to retrieve.
    folder: The local or Cloud Storage folder under which the files reside.
    tmap: The TensorMap to use for ECG input.

  Returns:
    An Altair plot or a notebook-friendly error.
  """
  tidy_resting_ecg_signal = reshape_resting_ecg_to_tidy(sample_id, folder, tmap)
  if tidy_resting_ecg_signal.shape[0] == 0:
    return HTML(f'''
      <div class="alert alert-block alert-danger">
      <b>Warning:</b> Resting ECG not available for sample {sample_id}.<br>
      Use the <kbd>folder</kbd> parameter to read HD5s from a different local directory or Cloud Storage bucket.
      </div>''')

  data_file = os.path.basename(RESTING_ECG_DATA_FILE.name)
  tidy_resting_ecg_signal.query("filtering in ['raw_mV']").to_json(data_file, orient='records')

  # Define the plot components.
  brush = alt.selection(type='interval', encodings=['x'])

  lead_dropdown = alt.binding_select(options=list(tidy_resting_ecg_signal.lead.unique()))
  lead_select = alt.selection_single(
      fields=['lead'], bind=lead_dropdown,
      name='Choose just one to view',
      init={'lead': tidy_resting_ecg_signal.lead.unique()[0]},
  )

  base = alt.Chart(data_file).mark_line().encode(
      x='ts_reference:Q',
      y='signal_mV:Q',
      color=alt.Color(
          'lead:N', legend=alt.Legend(orient='top'),
          title='Lead(s) currently displayed',
      ),
  ).properties(
      width=900, height=250, title=f'Resting ECG for {sample_id}',
  ).add_selection(
      lead_select,
  ).transform_filter(lead_select)

  upper = base.encode(x=alt.X('ts_reference:Q', scale=alt.Scale(domain=brush)))

  lower = base.properties(
      height=50, title='Brush over this subplot to select a time interval.',
  ).add_selection(brush)

  return upper & lower


def exercise_ecg_interactive_plot(
    sample_id: Union[int, str], folder: Optional[str] = None, time_interval_seconds: int = 10,
) -> Union[HTML, alt.Chart]:
  """Wrangle exercise ECG data to tidy and present it as an interactive plot.

  Args:
    sample_id: The id of the ECG sample to retrieve.
    folder: The local or Cloud Storage folder under which the files reside.
    time_interval_seconds: the width of the time interval (in seconds) to display of signal data

  Returns:
    An Altair plot or a notebook-friendly error.
  """
  (exercise_ecg_trend, exercise_ecg_signal) = reshape_exercise_ecg_to_tidy(sample_id=sample_id, folder=folder)
  if(exercise_ecg_trend.shape[0] == 0 or exercise_ecg_signal.shape[0] == 0):
    return HTML(f'''
      <div class="alert alert-block alert-danger">
      <b>Warning:</b> Exercise ECG not available for sample {sample_id}.<br>
      Use the <kbd>folder</kbd> parameter to read HD5s from a different local directory or Cloud Storage bucket.
      </div>''')

  trend_data_file = os.path.basename(EXERCISE_ECG_TREND_DATA_FILE.name)
  exercise_ecg_trend.to_json(trend_data_file, orient='records')
  signal_data_file = os.path.basename(EXERCISE_ECG_SIGNAL_DATA_FILE.name)
  exercise_ecg_signal.to_json(signal_data_file, orient='records')

  brush = alt.selection_single(on='mouseover', nearest=True, fields=['time'], init={'time': 200.0})

  lead_dropdown = alt.binding_select(options=list(exercise_ecg_signal.lead.unique()))
  lead_select = alt.selection_single(
      fields=['lead'], bind=lead_dropdown, name='Choose just one to view',
      init={'lead': exercise_ecg_signal.lead.unique()[0]},
  )

  trend = alt.Chart(trend_data_file).mark_point(opacity=0.8, filled=True, size=100).encode(
      x='time:Q',
      color=alt.Color('phasename:N', legend=alt.Legend(orient='top'), title='Phase names'),
      tooltip=[
          'artifact:Q', 'grade:Q', 'heartrate:Q', 'load:Q', 'mets:Q', 'pacecount:Q',
          'phasename:N', 'phasetime:Q', 'time:Q', 'vecount:Q',
      ],
  ).properties(
      width=900, height=100, title=f'Click on a point to select a {time_interval_seconds} second time interval.',
  ).add_selection(brush)

  signal = alt.Chart(signal_data_file).mark_line().encode(
      alt.X('time:Q', axis=alt.Axis(labelAngle=15)),
      y='raw_mV:Q',
      color=alt.Color('lead:N', legend=alt.Legend(orient='top'), title='Lead names'),
  ).properties(
      width=900, height=300, title='Exercise ECG signal for {}'.format(sample_id),
  ).add_selection(
      lead_select,
  ).transform_filter(
      lead_select,
  ).transform_filter(
      # https://github.com/altair-viz/altair/issues/1960
      f'''((toNumber({brush.name}.time) - {time_interval_seconds/2.0}) < datum.time)
           && (datum.time < toNumber({brush.name}.time) + {time_interval_seconds/2.0})''',
  )

  return trend.encode(y='heartrate:Q') & trend.encode(y='load:Q') & signal



"""Methods for Facets notebook integration."""
import base64
import os

import pandas as pd
from facets_overview.generic_feature_statistics_generator import GenericFeatureStatisticsGenerator

FACETS_DEPENDENCIES = {
    'facets_html': 'https://raw.githubusercontent.com/PAIR-code/facets/master/facets-dist/facets-jupyter.html',
    'webcomponents_js': 'https://cdnjs.cloudflare.com/ajax/libs/webcomponentsjs/1.3.3/webcomponents-lite.js',
}

if 'GOOGLE_PROJECT' in os.environ:  # This is Terra.
  # Terra notebook Content Security Policy prohibits pulling these files from
  # a remote location, so this code depends on the fact we can refer to it
  # from a location relative to the notebook.
  for dep, url in FACETS_DEPENDENCIES.items():
    if not os.path.exists(os.path.basename(url)):
      # If this is our Terra Docker image, the file is available locally.
      docker_image_path = os.path.join(os.environ['JUPYTER_HOME'], 'custom', os.path.basename(url))
      if os.path.exists(docker_image_path):
        os.system('ln -s ' + docker_image_path)
      else:
        os.system('wget --no-clobber ' + url)
    # Update dictionary to replace absolute url with relative url.
    FACETS_DEPENDENCIES[dep] = os.path.basename(url)


class FacetsOverview():
  """Methods for Facets Overview notebook integration."""

  def __init__(self, data: pd.DataFrame):
    # This takes the dataframe and computes all the inputs to the Facets
    # Overview plots such as:
    # - numeric variables: histogram bins, mean, min, median, max, etc..
    # - categorical variables: num unique, counts per category for bar chart,
    #     top category, etc.
    gfsg = GenericFeatureStatisticsGenerator()
    self._proto = gfsg.ProtoFromDataFrames(
        [{'name': 'data', 'table': data}],
    )

  def _repr_html_(self) -> str:
    """Html representation of Facets Overview for use in a Jupyter notebook."""
    protostr = base64.b64encode(self._proto.SerializeToString()).decode('utf-8')
    html_template = '''
        <script src="{webcomponents_js}"></script>
        <link rel="import" href="{facets_html}">
        <facets-overview id="overview_elem"></facets-overview>
        <script>
          document.querySelector("#overview_elem").protoInput = "{protostr}";
        </script>'''
    html = html_template.format(
        facets_html=FACETS_DEPENDENCIES['facets_html'],
        webcomponents_js=FACETS_DEPENDENCIES['webcomponents_js'],
        protostr=protostr,
    )
    return html


class FacetsDive():
  """Methods for Facets Dive notebook integration."""

  def __init__(self, data: pd.DataFrame, height: int = 1000):
    self._data = data
    self.height = height

  def _repr_html_(self) -> str:
    """Html representation of Facets Dive for use in a Jupyter notebook."""
    html_template = """
        <script src="{webcomponents_js}"></script>
        <link rel="import" href="{facets_html}">
        <facets-dive id="dive_elem" height="{height}"></facets-dive>
        <script>
          document.querySelector("#dive_elem").data = {data};
        </script>"""
    html = html_template.format(
        facets_html=FACETS_DEPENDENCIES['facets_html'],
        webcomponents_js=FACETS_DEPENDENCIES['webcomponents_js'],
        data=self._data.to_json(orient='records'),
        height=self.height,
    )
    return html



import h5py
import numpy as np
import datetime
from ml4h.TensorMap import TensorMap, Interpretation, str2date
from ml4h.defines import StorageType
from ml4h.metrics import weighted_crossentropy
from ml4h.tensormap.ukb.demographics import prevalent_tensor

DAYS_IN_5_YEARS = 365 * 5
DAYS_IN_10_YEARS = 365 * 10


def _survival_tensor(
    start_date_key: str,
    day_window: int,
    disease_name_override: str = None,
    incidence_only: bool = False,
    start_date_is_attribute: bool = False,
):
    def _survival_tensor_from_file(
        tm: TensorMap,
        hd5: h5py.File,
        dependents=None,
    ):
        if start_date_is_attribute:
            assess_date = datetime.datetime.utcfromtimestamp(hd5[start_date_key].attrs['date']).date()
        else:
            assess_date = str2date(str(hd5[start_date_key][0]))
        has_disease = 0  # Assume no disease if the tensor does not have the dataset

        if disease_name_override is not None:
            disease_name = disease_name_override
        else:
            disease_name = tm.name
        if disease_name in hd5['categorical']:
            has_disease = int(hd5['categorical'][disease_name][0])

        if disease_name + '_date' in hd5['dates']:
            censor_date = str2date(str(hd5['dates'][disease_name + '_date'][0]))
        elif 'phenotype_censor' in hd5['dates']:
            censor_date = str2date(str(hd5['dates/phenotype_censor'][0]))
        else:
            raise ValueError(f'No date found for survival {disease_name}')

        intervals = int(tm.shape[0] / 2)
        days_per_interval = day_window / intervals
        survival_then_censor = np.zeros(tm.shape, dtype=np.float32)
        for i, day_delta in enumerate(
                np.arange(0, day_window, days_per_interval),
        ):
            cur_date = assess_date + datetime.timedelta(days=day_delta)
            survival_then_censor[i] = float(cur_date < censor_date)
            survival_then_censor[intervals + i] = has_disease * float(
                censor_date <= cur_date < censor_date +
                datetime.timedelta(days=days_per_interval),
            )
            if i == 0 and censor_date <= cur_date:  # Handle prevalent diseases
                if incidence_only:
                    raise ValueError(f'{tm.name} ignores prior diagnoses.')
                survival_then_censor[intervals] = has_disease
        return survival_then_censor

    return _survival_tensor_from_file


def cox_tensor_from_file(
        start_date_key: str,
        incidence_only: bool = False,
        start_date_is_attribute: bool = False,
):
    def _cox_tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        if start_date_is_attribute:
            assess_date = datetime.datetime.utcfromtimestamp(hd5[start_date_key].attrs['date']).date()
        else:
            assess_date = str2date(str(hd5[start_date_key][0]))
        has_disease = 0  # Assume no disease if the tensor does not have the dataset
        if tm.name in hd5['categorical']:
            has_disease = int(hd5['categorical'][tm.name][0])

        if tm.name + '_date' in hd5['dates']:
            censor_date = str2date(str(hd5['dates'][tm.name + '_date'][0]))
        elif 'phenotype_censor' in hd5['dates']:
            censor_date = str2date(str(hd5['dates/phenotype_censor'][0]))
        else:
            raise ValueError(f'No date found for survival {tm.name}')

        if incidence_only and censor_date <= assess_date:
            raise ValueError(f'{tm.name} only considers incident diagnoses')

        tensor = np.zeros(tm.shape, dtype=np.float32)
        tensor[0] = has_disease
        tensor[1] = (censor_date - assess_date).days
        return tensor

    return _cox_tensor_from_file


enroll_cad_hazard = TensorMap(
    'coronary_artery_disease',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)
enroll_hyp_hazard = TensorMap(
    'hypertension',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)
enroll_afib_hazard = TensorMap(
    'atrial_fibrillation_or_flutter',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)

enroll_chol_hazard = TensorMap(
    'hypercholesterolemia',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)
enroll_diabetes2_hazard = TensorMap(
    'diabetes_type_2',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)
enroll_diabetes2_hazard_incident = TensorMap(
    'diabetes_type_2',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'dates/enroll_date',
        DAYS_IN_5_YEARS,
        incidence_only=True,
    ),
)
enroll_hyp_hazard_5 = TensorMap(
    'hypertension',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)
enroll_hyp_hazard_5_incident = TensorMap(
    'hypertension',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    tensor_from_file=_survival_tensor(
        'dates/enroll_date',
        DAYS_IN_5_YEARS,
        incidence_only=True,
    ),
)
enroll_cad_hazard_5_incident = TensorMap(
    'coronary_artery_disease_soft',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'dates/enroll_date',
        DAYS_IN_5_YEARS,
        incidence_only=True,
    ),
)
enroll_cad_hazard_5 = TensorMap(
    'coronary_artery_disease_soft',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)
enroll_mi_hazard_5 = TensorMap(
    'myocardial_infarction',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor('dates/enroll_date', DAYS_IN_5_YEARS),
)
enroll_mi_hazard_5_incident = TensorMap(
    'myocardial_infarction',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'dates/enroll_date',
        DAYS_IN_5_YEARS,
        incidence_only=True,
    ),
)

cox_mi = TensorMap(
    'myocardial_infarction',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file('dates/enroll_date'),
)
cox_mi_incident = TensorMap(
    'myocardial_infarction',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file(
        'dates/enroll_date', incidence_only=True,
    ),
)
cox_hyp = TensorMap(
    'hypertension',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file('dates/enroll_date'),
)
cox_hyp_incident = TensorMap(
    'hypertension',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file(
        'dates/enroll_date', incidence_only=True,
    ),
)
cox_cad = TensorMap(
    'coronary_artery_disease_soft',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file('dates/enroll_date'),
)
cox_cad_incident = TensorMap(
    'coronary_artery_disease_soft',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file(
        'dates/enroll_date', incidence_only=True,
    ),
)
cox_cad = TensorMap(
    'coronary_artery_disease_soft',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file('dates/enroll_date'),
)
cox_cad_incident = TensorMap(
    'coronary_artery_disease_soft',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file(
        'dates/enroll_date', incidence_only=True,
    ),
)

cox_afib_wrt_instance2 = TensorMap(
    'atrial_fibrillation_or_flutter',
    Interpretation.TIME_TO_EVENT,
    tensor_from_file=cox_tensor_from_file(
        'ukb_cardiac_mri/cine_segmented_lax_2ch/2/instance_0/', start_date_is_attribute=True,
    ),
)

survival_afib_wrt_instance2 = TensorMap(
    'atrial_fibrillation_or_flutter',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_cardiac_mri/cine_segmented_lax_2ch/2/instance_0/',
        DAYS_IN_5_YEARS, start_date_is_attribute=True,
    ),
)

mgb_afib_wrt_instance2 = TensorMap(
    'survival_curve_af',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_5_YEARS,
        disease_name_override='atrial_fibrillation_or_flutter',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_af_wrt_instance2 = TensorMap(
    'af_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_5_YEARS,
        disease_name_override='atrial_fibrillation_or_flutter',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_cad_wrt_instance2 = TensorMap(
    'cad_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_5_YEARS,
        disease_name_override='coronary_artery_disease',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_dm_wrt_instance2 = TensorMap(
    'dm_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_5_YEARS,
        disease_name_override='diabetes_type_2',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_mi_wrt_instance2 = TensorMap(
    'mi_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_5_YEARS,
        disease_name_override='myocardial_infarction',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_hf_wrt_instance2 = TensorMap(
    'hf_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_10_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_10_YEARS,
        disease_name_override='heart_failure',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_hf_wrt_instance2 = TensorMap(
    'hf_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_10_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_10_YEARS,
        disease_name_override='heart_failure',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_hf_nlp_wrt_instance2 = TensorMap(
    'hf_nlp_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_10_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_10_YEARS,
        disease_name_override='heart_failure',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_hf_primary_wrt_instance2 = TensorMap(
    'hf_primary_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_10_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_10_YEARS,
        disease_name_override='heart_failure',
        start_date_is_attribute=True, incidence_only=True,
    ),
)

mgb_hf_death_wrt_instance2 = TensorMap(
    'hf_death_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_10_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_10_YEARS,
        disease_name_override='death',
        start_date_is_attribute=True, incidence_only=True,
    ),
)

mgb_death_wrt_instance2 = TensorMap(
    'death_event',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_10_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_10_YEARS,
        disease_name_override='death',
        start_date_is_attribute=True, incidence_only=True,
    ),
)
mgb_afib_as_hf_wrt_instance2 = TensorMap(
    'survival_curve_af',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_5_YEARS,
        disease_name_override='heart_failure',
        start_date_is_attribute=True, incidence_only=True,
    ),
)

mgb_afib_as_stroke_wrt_instance2 = TensorMap(
    'survival_curve_af',
    Interpretation.SURVIVAL_CURVE,
    shape=(50,),
    days_window=DAYS_IN_5_YEARS,
    tensor_from_file=_survival_tensor(
        'ukb_ecg_rest/ecg_rest_text/instance_0', DAYS_IN_5_YEARS,
        disease_name_override='stroke',
        start_date_is_attribute=True, incidence_only=True,
    ),
)


prevalent_hf_wrt_instance2 = TensorMap(
    'heart_failure', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    loss=weighted_crossentropy([1.0, 58], 'heart_failure'), path_prefix='categorical',
    channel_map={'no_heart_failure': 0, 'prevalent_heart_failure': 1},
    tensor_from_file=prevalent_tensor(
        'ukb_cardiac_mri/cine_segmented_lax_2ch/2/instance_0/',
        'dates/heart_failure_date',
        start_date_is_attribute=True,
    ),

)

prevalent_af_as_hf_wrt_instance2 = TensorMap(
    'heart_failure', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    loss=weighted_crossentropy([1.0, 58], 'heart_failure'), path_prefix='categorical',
    channel_map={'no_atrial_fibrillation_or_flutter': 0, 'prevalent_atrial_fibrillation_or_flutter': 1},
    tensor_from_file=prevalent_tensor(
        'ukb_cardiac_mri/cine_segmented_lax_2ch/2/instance_0/',
        'dates/atrial_fibrillation_or_flutter_date',
        start_date_is_attribute=True,
    ),

)

prevalent_cad_wrt_instance2 = TensorMap(
    'coronary_artery_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    loss=weighted_crossentropy([1.0, 10], 'coronary_artery_disease'), path_prefix='categorical',
    channel_map={'no_coronary_artery_disease': 0, 'prevalent_coronary_artery_disease': 1},
    tensor_from_file=prevalent_tensor(
        'ukb_cardiac_mri/cine_segmented_lax_2ch/2/instance_0/',
        'dates/coronary_artery_disease_date',
        start_date_is_attribute=True,
    ),

)



"""Methods for reshaping raw ECG signal data for use in the pandas ecosystem."""
import os
import tempfile
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd
from biosppy.signals.tools import filter_signal
import h5py
from ml4h.defines import ECG_BIKE_LEADS
from ml4h.defines import ECG_REST_LEADS
from ml4h.runtime_data_defines import get_exercise_ecg_hd5_folder
from ml4h.runtime_data_defines import get_resting_ecg_hd5_folder
from ml4h.TensorMap import TensorMap
import ml4h.tensormap.ukb.ecg as ecg_tmaps
import tensorflow as tf


RAW_SCALE = 0.005  # Convert to mV.
SAMPLING_RATE = 500.0
DEFAULT_RESTING_ECG_SIGNAL_TMAP = ecg_tmaps.ecg_rest

# TODO(deflaux): parameterize exercise ECG by TMAP name if there is similar ECG data from other studies.
EXERCISE_ECG_SIGNAL_TMAP = ecg_tmaps.ecg_bike_raw_full
EXERCISE_ECG_TREND_TMAPS = [
    ecg_tmaps.ecg_bike_raw_trend_hr,
    ecg_tmaps.ecg_bike_raw_trend_load,
    ecg_tmaps.ecg_bike_raw_trend_grade,
    ecg_tmaps.ecg_bike_raw_trend_artifact,
    ecg_tmaps.ecg_bike_raw_trend_mets,
    ecg_tmaps.ecg_bike_raw_trend_pacecount,
    ecg_tmaps.ecg_bike_raw_trend_phasename,
    ecg_tmaps.ecg_bike_raw_trend_phasetime,
    ecg_tmaps.ecg_bike_raw_trend_time,
    ecg_tmaps.ecg_bike_raw_trend_vecount,
]
EXERCISE_PHASES = {0.0: 'Pretest', 1.0: 'Exercise', 2.0: 'Recovery'}


def _examine_available_keys(hd5: Dict[str, Any]) -> None:
  print(f'hd5 ECG keys {[k for k in hd5.keys() if "ecg" in k]}')
  for key in [k for k in hd5.keys() if 'ecg' in k]:
    print(f'hd5 {key} keys {k for k in hd5[key]}')


def reshape_resting_ecg_to_tidy(
    sample_id: Union[int, str], folder: Optional[str] = None, tmap: TensorMap = DEFAULT_RESTING_ECG_SIGNAL_TMAP,
) -> pd.DataFrame:
  """Wrangle resting ECG data to tidy.

  Args:
    sample_id: The id of the ECG sample to retrieve.
    folder: The local or Cloud Storage folder under which the files reside.
    tmap: The TensorMap to use for ECG input.

  Returns:
    A pandas dataframe in tidy format or print a notebook-friendly error and return an empty dataframe.
  """
  if folder is None:
    folder = get_resting_ecg_hd5_folder(sample_id)

  data: Dict[str, Any] = {'lead': [], 'raw': [], 'ts_reference': [], 'filtered': [], 'filtered_1': [], 'filtered_2': []}

  with tempfile.TemporaryDirectory() as tmpdirname:
    sample_hd5 = str(sample_id) + '.hd5'
    local_path = os.path.join(tmpdirname, sample_hd5)
    try:
      tf.io.gfile.copy(src=os.path.join(folder, sample_hd5), dst=local_path)
    except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
      print(f'''Warning: Resting ECG not available for sample {sample_id} in folder {folder}.
      Use the folder parameter to read HD5s from a different directory or bucket.\n\n{e.message}''')
      return pd.DataFrame(data)

    with h5py.File(local_path, mode='r') as hd5:
      try:
        signals = tmap.tensor_from_file(tmap, hd5)
      except (KeyError, ValueError) as e:
        print(f'''Warning: Resting ECG TMAP {tmap.name} not available for sample {sample_id}.
        Use the tmap parameter to choose a different TMAP.\n\n{e}''')
        _examine_available_keys(hd5)
        return pd.DataFrame(data)
      for (lead, channel) in ECG_REST_LEADS.items():
        signal = signals[:, channel]
        signal_length = len(signal)
        data['raw'].extend(signal)
        data['lead'].extend([lead] * signal_length)
        data['ts_reference'].extend(np.array([i*1./(SAMPLING_RATE+1.) for i in range(0, signal_length)]))
        filtered, _, _ = filter_signal(
            signal=signal,
            ftype='FIR',
            band='bandpass',
            order=int(0.3 * SAMPLING_RATE),
            frequency=[.9, 50],
            sampling_rate=SAMPLING_RATE,
        )
        data['filtered'].extend(filtered)
        filtered_1, _, _ = filter_signal(
            signal=signal,
            ftype='FIR',
            band='bandpass',
            order=int(0.3 * SAMPLING_RATE),
            frequency=[.9, 20],
            sampling_rate=SAMPLING_RATE,
        )
        data['filtered_1'].extend(filtered_1)
        filtered_2, _, _ = filter_signal(
            signal=signal,
            ftype='FIR',
            band='bandpass',
            order=int(0.3 * SAMPLING_RATE),
            frequency=[.9, 30],
            sampling_rate=SAMPLING_RATE,
        )
        data['filtered_2'].extend(filtered_2)

  signal_df = pd.DataFrame(data)
  # Convert the raw signal to mV.
  signal_df['raw_mV'] = signal_df['raw'] * RAW_SCALE
  signal_df['filtered_mV'] = signal_df['filtered'] * RAW_SCALE
  signal_df['filtered_1_mV'] = signal_df['filtered_1'] * RAW_SCALE
  signal_df['filtered_2_mV'] = signal_df['filtered_2'] * RAW_SCALE
  # Reshape to tidy (long format).
  tidy_signal_df = signal_df.melt(
      id_vars=['lead', 'ts_reference'],
      value_vars=['raw_mV', 'filtered_mV', 'filtered_1_mV', 'filtered_2_mV'],
      var_name='filtering', value_name='signal_mV',
  )

  # The leads have a meaningful order, apply the order to this column.
  lead_factor_type = pd.api.types.CategoricalDtype(
      categories=[
          'strip_I', 'strip_aVR', 'strip_V1', 'strip_V4',
          'strip_II', 'strip_aVL', 'strip_V2', 'strip_V5',
          'strip_III', 'strip_aVF', 'strip_V3', 'strip_V6',
      ],
      ordered=True,
  )
  tidy_signal_df['lead'] = tidy_signal_df.lead.astype(lead_factor_type)

  return tidy_signal_df


def reshape_exercise_ecg_to_tidy(
    sample_id: Union[int, str], folder: Optional[str] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
  """Wrangle exercise ECG signal data to tidy format.

  Args:
    sample_id: The id of the ECG sample to retrieve.
    folder: The local or Cloud Storage folder under which the files reside.

  Returns:
    A tuple of pandas dataframes or print a notebook-friendly error and return empty dataframes.
    * first tuple element is trend data in wide format
    * second tuple element is signal data in tidy format
  """
  if folder is None:
    folder = get_exercise_ecg_hd5_folder(sample_id)

  with tempfile.TemporaryDirectory() as tmpdirname:
    sample_hd5 = str(sample_id) + '.hd5'
    local_path = os.path.join(tmpdirname, sample_hd5)
    try:
      tf.io.gfile.copy(src=os.path.join(folder, sample_hd5), dst=local_path)
    except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
      print(f'''Warning: Exercise ECG not available for sample {sample_id} in folder {folder}.
      Use the folder parameter to read HD5s from a different directory or bucket.\n\n{e.message}''')
      return (pd.DataFrame({}), pd.DataFrame({}))

    with h5py.File(local_path, mode='r') as hd5:
      trend_data = {}
      for tmap in EXERCISE_ECG_TREND_TMAPS:
        try:
          tensor = tmap.tensor_from_file(tmap, hd5)
          trend_data[tmap.name.replace('trend_', '')] = tensor
        except (KeyError, ValueError) as e:
          print(f'Warning: Exercise ECG trend not available for sample {sample_id}\n\n{e}')
          _examine_available_keys(hd5)
          return (pd.DataFrame({}), pd.DataFrame({}))
      try:
        full = EXERCISE_ECG_SIGNAL_TMAP.tensor_from_file(EXERCISE_ECG_SIGNAL_TMAP, hd5)
      except (KeyError, ValueError) as e:
        print(f'Warning: Exercise ECG not available for sample {sample_id}\n\n{e}')
        _examine_available_keys(hd5)
        return (pd.DataFrame({}), pd.DataFrame({}))

  signal_data = {}
  for (lead, channel) in ECG_BIKE_LEADS.items():
    signal_data['raw_mV_' + lead] = full[:, channel] * RAW_SCALE
  signal_data['time'] = np.arange(len(full)) / SAMPLING_RATE

  # Convert exercise ecg trend tensor dictionarys to a dataframe. Include only trends with the same length
  # as the 'time' trend. (note: this appears to only be a problem with the fake data.)
  trend_df = pd.DataFrame({k: v for k, v in trend_data.items() if v.shape[0] == trend_data['time'].shape[0]})
  # Clean data - convert to categorical string.
  trend_df['phasename'] = trend_df.phasename.map(EXERCISE_PHASES).astype('category')

  # Convert exercise ecg signal tensor dictionary to a dataframe, clean data
  # as needed, and then pivot to tidy.
  signal_df = pd.DataFrame(signal_data)
  tidy_signal_df = pd.wide_to_long(
      signal_df,
      stubnames=['raw_mV'],
      i='time',
      j='lead',
      sep='_',
      suffix='.*',
  )
  tidy_signal_df.reset_index(inplace=True)  # Turn pd multiindex into columns.
  # The leads have a meaningful order, apply the order to this column.
  lead_factor_type = pd.api.types.CategoricalDtype(categories=ECG_BIKE_LEADS.keys(), ordered=True)
  tidy_signal_df['lead'] = tidy_signal_df.lead.astype(lead_factor_type)

  return (trend_df, tidy_signal_df)


def reshape_exercise_ecg_and_trend_to_tidy(
    sample_id: Union[int, str], folder: Optional[str] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
  """Wrangle exercise ECG signal and trend data to tidy format.

  Args:
    sample_id: The id of the ECG sample to retrieve.
    folder: The local or Cloud Storage folder under which the files reside.

  Returns:
    A tuple of pandas dataframes or print a notebook-friendly error and return empty dataframes.
    * first tuple element is trend data in tidy format
    * second tuple element is signal data in tidy format
  """

  # Get the trend data in wide format and pivot it to tidy.
  (trend_df, tidy_signal_df) = reshape_exercise_ecg_to_tidy(sample_id, folder)
  # Clean data - drop zero-valued columns.
  trend_df = trend_df.loc[:, ~trend_df.eq(0).all()]
  trend_id_vars = ['time', 'phasename', 'phasetime']
  trend_value_vars = trend_df.columns[~trend_df.columns.isin(trend_id_vars)].tolist()
  tidy_trend_df = trend_df.melt(
      id_vars=trend_id_vars,
      value_vars=trend_value_vars,
      var_name='measurement',
      value_name='value',
  )

  return (tidy_trend_df, tidy_signal_df)
